/*
 * modversions - A simple utility to display module versioning information
 *               from Linux Kernel Modules (.ko ELF objects).
 */

#include <stdio.h>
#include <elf.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

static int elf_endian;
static int my_endian;

static inline void __endian(const void *src, void *dest, unsigned int size)
{
	unsigned int i;
	for (i = 0; i < size; i++)
	((unsigned char*)dest)[i] = ((unsigned char*)src)[size - i-1];
}

#define TO_NATIVE(x)                                                      \
({                                                                        \
	typeof(x) __x;                                                    \
	if (elf_endian != my_endian) __endian(&(x), &(__x), sizeof(__x)); \
	else __x = x;                                                     \
	__x;                                                              \
})

struct modver32_info
{
	uint32_t crc;
	char name[64 - sizeof(uint32_t)];
};

struct modver64_info
{
       uint64_t crc;
       char name[64 - sizeof(uint64_t)];
};

const char *skip_dot(const char *str)
{
       /* For our purposes, .foo matches foo.  PPC64 needs this. */
       if (str && str[0] == '.')
               return str + 1;
       return str;
}

static int elf_ident(void *ko, unsigned long size)
{
	/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
	char *ident = ko;

	if (size < EI_CLASS || memcmp(ko, ELFMAG, SELFMAG) != 0)
		return ELFCLASSNONE;
	elf_endian = ident[EI_DATA];
	return ident[EI_CLASS];
}

static void *get_section32(void *file,
			   unsigned long size,
			   const char *name,
			   unsigned long *secsize)
{
	Elf32_Ehdr *elf_header = file;
	Elf32_Shdr *elf_sections = file + TO_NATIVE(elf_header->e_shoff);
	char *section_names = NULL;
	unsigned int i;

	if (size < sizeof(*elf_header))
		return NULL;
	if (size < elf_header->e_shoff +
	    TO_NATIVE(elf_header->e_shnum) * sizeof(elf_sections[0]))
		return NULL;
	if (size < TO_NATIVE(elf_sections[TO_NATIVE(elf_header->e_shstrndx)].sh_offset))
		return NULL;

	section_names = file
		+ TO_NATIVE(elf_sections[TO_NATIVE(elf_header->e_shstrndx)].sh_offset);
	for (i = 1; i < TO_NATIVE(elf_header->e_shnum); i++) {
		if (0 == (strncmp(name,
			  section_names+TO_NATIVE(elf_sections[i].sh_name),
			  strlen(section_names+TO_NATIVE(elf_sections[i].sh_name))))) {
				*secsize = TO_NATIVE(elf_sections[i].sh_size);
				return file + TO_NATIVE(elf_sections[i].sh_offset);
		}
	}

	return NULL;
}

static void *get_section64(void *file,
			   unsigned long size,
			   const char *name,
			   unsigned long *secsize)
{
	Elf64_Ehdr *elf_header = file;
	Elf64_Shdr *elf_sections = file + TO_NATIVE(elf_header->e_shoff);
	char *section_names = NULL;
	unsigned int i;

	if (size < sizeof(*elf_header))
		return NULL;
	if (size < TO_NATIVE(elf_header->e_shoff) +
	    TO_NATIVE(elf_header->e_shnum) * sizeof(elf_sections[0]))
		return NULL;
	if (size < TO_NATIVE(elf_sections[TO_NATIVE(elf_header->e_shstrndx)].sh_offset))
		return NULL;

	section_names = file
		+ TO_NATIVE(elf_sections[TO_NATIVE(elf_header->e_shstrndx)].sh_offset);
	for (i = 1; i < TO_NATIVE(elf_header->e_shnum); i++) {
		if (0 == (strncmp(name,
			  section_names+TO_NATIVE(elf_sections[i].sh_name),
			  strlen(section_names+TO_NATIVE(elf_sections[i].sh_name))))) {
				*secsize = TO_NATIVE(elf_sections[i].sh_size);
				return file + TO_NATIVE(elf_sections[i].sh_offset);
		}
	}

	return NULL;
}

static void *get_section(void *ko,
			 unsigned long size,
			 const char *name,
			 unsigned long *secsize)
{
	switch (elf_ident(ko, size)) {
	case ELFCLASS32:
		return get_section32(ko, size, name, secsize);
	case ELFCLASS64:
		return get_section64(ko, size, name, secsize);
	default:
		return NULL;
	}
}

void *get_module(const char *name, unsigned long *size)
{

	int fd = -1;
	void *ko = NULL;
	struct stat st;

	if (0 > (fd = open(name, O_RDONLY))) {
		perror("cannot open module");
		exit(1);
	}

	fstat(fd, &st);
	*size = st.st_size;
	ko = mmap(0, *size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (MAP_FAILED == ko) {
		perror("cannot map open module file");
		exit(1);
	}

	return ko;
}

int main(int argc, char **argv)
{

	int i = -1;
	unsigned long size,secsize = 0;
	struct modver32_info *info32 = NULL;
	struct modver64_info *info64 = NULL;
	void *ko = NULL;
	char *filename = NULL;

	union { short s; char c[2]; } endian_test;

	endian_test.s = 1;
	if (endian_test.c[1] == 1) my_endian = ELFDATA2MSB;
	else if (endian_test.c[0] == 1) my_endian = ELFDATA2LSB;
	else {
		perror("NUXI overload detected");
		exit(1);
	}

	if (2 == argc)
		filename = argv[1];
	else {
		printf("%s: unable to open module file\n", argv[0]);
		exit(1);
	}

	ko = get_module(filename, &size);

	switch(elf_ident(ko, size)) {
	case ELFCLASS32:
		info32 = get_section(ko, size, "__versions", &secsize);
		for (i = 0; i < secsize / sizeof(struct modver32_info); i++) {
			printf("0x%08lx\t%s\n", (unsigned long)
				info32[i].crc, skip_dot(info32[i].name));
		}
		break;
	case ELFCLASS64:
		/* For compatibility, only pad 8 (and not 6) zeros. */
		info64 = get_section(ko, size, "__versions", &secsize);
		for (i = 0; i < secsize / sizeof(struct modver64_info); i++) {
			printf("0x%08llx\t%s\n", (unsigned long long)
				info64[i].crc, skip_dot(info64[i].name));
		}
		break;
	default:
		perror("cannot identify module");
		exit(1);
	}
	return 0;
}
